"""The controller used for all high level management of the gacha game. 
Instances of this class should be created for each instance of a gacha game 
you would like to run. These instances should then be used to access and 
manage the Item, Banner, and Player objects in the game directly. Once 
accessed, the objects can then be modified directly using class methods (see 
objects.py for more info).

This object operates as the Controller of a gachapy game 
"""

from typing import Dict, Optional
from gachapy.objects import *


def default_key(rarity: float) -> float:
    """The default function that converts rarity to drop rate (1/x)

    Parameters
    ----------
    rarity : float
        the rarity of the item

    Returns
    -------
    float
        the drop rate of the item
    """
    return 1 / rarity


def sort_player_key(player: Player) -> int:
    """The key used to sort players in a list of players

    Parameters
    ----------
    player : Player
        the player to extract the key from

    Returns
    -------
    int
        the key of the player
    """
    return sum([i.rarity for i in player.items])


class PullError(Exception):
    """An exception thrown when pulling from a banner"""

    pass


class Controller:
    """A controller used to control an instance of a gacha game. Used to
    perform all actions needed in a game and manage game information in the
    form of gachapy objects

    Fields
    ------
    items : List[Item]
        the list of items that are in the item pool for the gacha
    banners : List[Banner]
        the list of banners that are available for the gacha
    players : List[Player]
        the list of players enrolled in the gacha
    """

    def __init__(
        self,
        items: Dict[str, Item] = {},
        banners: Dict[str, Banner] = {},
        players: Dict[str, Player] = {},
    ) -> None:
        """Creates an instance of a gacha controller

        Parameters
        ----------
        items : Dict[str,Item]
            the dict of id, item pairs that are in the item pool for the gacha
        banners : Dict[str,Banner]
            the dict of id, banner pairs that are available for the gacha
        players : Dict,str[Player]
            the dict of id, player pairs enrolled in the gacha
        """
        self.items = items
        self.banners = banners
        self.players = players

    def find_item_by_name(self, item_name: str) -> Optional[Item]:
        """Returns the Item object with the given name or None if not found

        Parameters
        ----------
        item_name : str
            the name of the item

        Returns
        -------
        Optional[Item]
            the item object with the given name or None if not found
        """
        item = [i for i in self.items.values() if i.name == item_name]
        if len(item) < 1:
            return None
        return item[0]

    def find_banner_by_name(self, banner_name: str) -> Optional[Banner]:
        """Returns the Banner object with the given name or None if not found

        Parameters
        ----------
        banner_name : str
            the name of the banner

        Returns
        -------
        Optional[Banner]
            the banner object with the given name or None if not found
        """
        banner = [i for i in self.banners.values() if i.name == banner_name]
        if len(banner) < 1:
            return None
        return banner[0]

    def find_player_by_name(self, player_name: str) -> Optional[Player]:
        """Returns the Player object with the given name or None if not found

        Parameters
        ----------
        player_name : str
            the name of the player

        Returns
        -------
        Optional[player]
            the player object with the given name or None if not found
        """
        player = [i for i in self.players.values() if i.name == player_name]
        if len(player) < 1:
            return None
        return player[0]

    def find_item_by_id(self, item_id: str) -> Optional[Item]:
        """Returns the Item object with the given id or None if not found

        Parameters
        ----------
        item_id : str
            the id of the item

        Returns
        -------
        Optional[Item]
            the item object with the given id or None if not found
        """
        return self.items.get(item_id)

    def find_banner_by_id(self, banner_id: str) -> Optional[Banner]:
        """Returns the Banner object with the given id or None if not found

        Parameters
        ----------
        banner_id : str
            the id of the banner

        Returns
        -------
        Optional[Banner]
            the banner object with the given id or None if not found
        """
        return self.banners.get(banner_id)

    def find_player_by_id(self, player_id: str) -> Optional[Player]:
        """Returns the Player object with the given id or None if not found

        Parameters
        ----------
        player_id : str
            the id of the player

        Returns
        -------
        Optional[player]
            the player object with the given id or None if not found
        """
        return self.players.get(player_id)

    def pull(self, player_id: str, banner_id: str) -> Optional[Item]:
        """Pulls and returns an item from the specified banner for the specified player

        Parameters
        ----------
        player_id : str
            the id of the selected player
            Preconditon: must be a valid id
        banner_id : str
            the id of the selected banner
            Preconditon: must be a valid id

        Returns
        -------
        Optional[Item]
            the item if the pull is successful or None if the player does not have enough money

        Raises
        ------
        PullError if player or banner are not valid
        """
        player = self.find_player_by_id(player_id)
        if player == None:
            raise PullError("Player not found")
        banner = self.find_banner_by_id(banner_id)
        if banner == None:
            raise PullError("Banner not found")
        if player.change_money(-1 * banner.price):
            item = banner.pull()
            player.add_item(item)
            return item
        return None

    def add_new_item(self, name: str, id: str, rarity: float) -> Optional[Item]:
        """Adds a new item to the gacha game

        Parameters
        ----------
        name : str
            the name of the new item
        id : str
            the id of the new item
            Precondition: must be unique
        rarity : float
            the rarity of the item

        Returns
        -------
        Optional[Item]
            the Item object representing the new item or None if an item
            with the specified id already exists
        """
        item = self.find_item_by_id(id)
        if item != None:
            return None
        new_item = Item(name, id, rarity)
        self.items[id] = new_item
        return new_item

    def add_new_banner(
        self,
        name: str,
        id: str,
        item_list_str: List[str],
        price: float,
        key: Callable[[int], float] = default_key,
    ) -> Optional[Banner]:
        """Adds a new banner to the gacha game

        Parameters
        ----------
        name : str
            the name of the new banner
        id : str
            the id of the new banner
            Precondition: must be unique
        item_list_str : List[str]
            the list of the ids of the items in the banner
        price : float
            the price of pulling from the banner
        key : function : int -> float
            function that takes in rarity and returns the drop rate of the item

        Returns
        -------
        Optional[Banner]
            the Banner object representing the new banner or None if a banner
            with the specified id already exists
        """
        banner = self.find_banner_by_name(name)
        if banner != None:
            return None
        item_list = [self.find_item_by_id(i) for i in item_list_str]
        new_banner = Banner(name, id, item_list, price, key)
        self.banners[id] = new_banner
        return new_banner

    def add_new_player(
        self, name: str, id: str, start_money: float, items_str: List[str] = []
    ) -> Optional[Player]:
        """Adds a new player to the gacha game

        Parameters
        ----------
        name : str
            the name of the new player
        id : str
            the id of the new player
        start_money : float
            the amount of money the new player will start with
        items_str : List[str]
            the list of the ids of the items the player has

        Returns
        -------
        Optional[Player]
            the Player object representing the new player or None if a player
            with the specified id already exists
        """
        player = self.find_player_by_id(id)
        if player != None:
            return None
        items_list = [self.find_item_by_id(i) for i in items_str]
        new_player = Player(name, id, items_list, start_money)
        self.players[id] = new_player
        return new_player

    def remove_item(self, item_id: str) -> Optional[Item]:
        """Removes the specified item from the gacha game
        WARNING: Will also remove from banners and players if found,
        costly operation

        Parameters
        ----------
        item_id : str
            the name of the item to remove

        Returns
        -------
        Optional[Item]
            the removed item or None if item does not exist
        """
        item = self.items.pop(item_id, None)
        if item == None:
            return item
        for banner in self.banners:
            for item in banner.items:
                if item.id == item_id:
                    banner.items.remove(item)
        for player in self.players:
            for item in player.items:
                if item.id == item_id:
                    player.items.remove(item)
        return item

    def remove_banner(self, banner_id: str) -> Optional[Banner]:
        """Removes the specified banner from the gacha game

        Parameters
        ----------
        banner_id : str
            the id of the banner to remove

        Returns
        -------
        Optional[Banner]
            the removed banner or None if banner does not exist
        """
        return self.banners.pop(banner_id, None)

    def remove_player(self, player_id: str) -> Optional[Player]:
        """Removes the specified player from the gacha game

        Parameters
        ----------
        player_id : str
            the id of the player to remove

        Returns
        -------
        Optional[Player]
            the removed player or None if player does not exist
        """
        return self.players.pop(player_id, None)

    def create_random_banner(
        self,
        name: str,
        id: str,
        num_items: int,
        price: float = -1,
        key: Callable[[int], float] = default_key,
    ) -> Optional[Banner]:
        """Creates a random banner with the given name and number of items
            The price is automatically determined by the average of the rarities of the items
            selected if a price is not specified

        Parameters
        ----------
        name : str
            the name of the random banner
        id : str
            the id of the random banner
            Precondition: must be unique
        num_items : int
            the number of items in the banner
        price : float
            the price of the banner
        key : function : int -> float
            function that takes in rarity and returns the drop rate of the item

        Returns
        -------
        Optional[Banner]
            the banner created or None if a banner with the name specified already exists
        """
        item_list = random.choices(list(self.items.values()), k=num_items)
        item_list_str = [item.id for item in item_list]
        if price < 0:
            price = 0
            for item in item_list:
                price += item.rarity
            price /= len(item_list)
        return self.add_new_banner(name, id, item_list_str, price, key)

    def remove_all_banners(self) -> None:
        """Removes all of the banners in the game

        Returns
        -------
        None
        """
        self.banners = {}

    def top_items(self, num_items: int) -> List[Item]:
        """Returns the top specified number of items in the game in terms of rarity

        Parameters
        ----------
        num_items : int
            the number of items to return

        Returns
        -------
        List[Item]
            the list of top items
        """
        sort_list = sorted(list(self.items.values()), key=sort_item_key, reverse=True)
        return sort_list[: num_items - 1]

    def top_players(self, num_players: int) -> List[Player]:
        """Returns the top specified number of players in the game in terms of net worth

        Parameters
        ----------
        num_players : int
            the number of players to return

        Returns
        -------
        List[Player]
            the list of top players
        """
        sort_list = sorted(
            list(self.players.values()), key=sort_player_key, reverse=True
        )
        return sort_list[: num_players - 1]



"""Objects used for low level management and data storage of the gacha game. 
Instances of these classes should be created indirectly through a Controller 
(see controller.py for more info) but can be modified directly through the use 
of class methods. DO NOT directly edit fields or the invariants could be 
disrupted (although viewing fields directly is permitted). 

These objects operate as the Model of a gachapy game 
"""
import random
from typing import Callable, List


class Item:
    """A representation of an item in the gacha game.
    An item is something that players can collect and pull from banners. Each
    item must be unique

    Fields
    ------
    name : string
        name of the item
    id : string
        description of the item
        Invariant: must be unique
    id : int
        rarity of the item where the higher the numer, the higher the rarity
        Invariant: must be >= 1
    """

    def __init__(self, name: str, id: str, rarity: float) -> None:
        """Creates an Item object

        Parameters
        ----------
        name : str
            name of the item
        id : str
            id of the item
            Precondition: must be unique
        rarity : float
            rarity of the item where the higher the numer, the higher the rarity
        """
        self.name = name
        self.id = id
        self.rarity = rarity

    def change_rarity(self, rarity: float) -> bool:
        """Changes the rarity of the Item

        Parameters
        ----------
        rarity : float
            new rarity of the item
            Precondition: rarity must be >= 1

        Returns
        -------
        bool
            True if the rarity successfully updated, false otherwise
        """
        if rarity < 1:
            return False
        self.rarity = rarity
        return True

    def __str__(self) -> str:
        """Returns a string representation of this Item object
        self.name (Rarity: self.rarity)

        Returns
        -------
        str
            String representation of this object
        """
        return self.name + "\nID: " + self.id + "\nRarity: " + str(self.rarity)


class Banner:
    """A representation of a banner in the gacha game.
    A Banner is where players can choose to spend money in order to obtain
    a random item from the specific item pool presented in the Banner. Each
    Banner (should) contain different items and have different prices than
    other banners to distinguish their offerings

    Fields
    ------
    name : str
        name of the banner
    id : str
        id of the banner
        Invariant: must be unique
    item_list : List[Item]
        the list of items in the banner
    price : float
        the price of pulling from the banner
    key : function : int -> float
        function that takes in rarity and returns the drop rate of the item
    weights : List[float]
        list of drop weights for items
        Invariant: weights[i] corresponds to item_list[i]
    """

    def __init__(
        self,
        name: str,
        id: str,
        item_list: List[Item],
        price: float,
        key: Callable[[float], float],
    ) -> None:
        """Creates a Banner object

        Parameters
        ----------
        name : str
            name of the banner
        id : str
            id of the banner
            Precondition: must be unique
        item_list : List[Item]
            the list of items in the banner
            Precondition: all items must be unique
        weights : List[float]
            the list of drop weights
            Precondition: weights[i] is the drop weight for item_list[i]
        price : float
            the price of pulling from the banner
        key : function : int -> float
            function that takes in rarity and returns the drop rate of the item
        """
        self.name = name
        self.id = id
        self.item_list = item_list
        self.key = key
        self.weights = get_random_weights(item_list, key)
        self.price = price

    def add_item(self, item: Item) -> None:
        """Adds an item to the banner

        Parameters
        ----------
        item : Item
            item to add to the banner

        Returns
        -------
            None
        """
        self.item_list.append(item)
        self.weights = get_random_weights(self.item_list)

    def remove_item(self, item: Item) -> bool:
        """Removes an item from the banner

        Parameters
        ----------
        item : Item
            item to remove from the banner

        Returns
        -------
        bool
            True if item is found in banner, False if otherwise
        """
        try:
            self.item_list.remove(item)
            self.weights = get_random_weights(self.item_list)
            return True
        except:
            return False

    def pull(self) -> Item:
        """Returns a random item out of a banner randomized by weight

        Returns
        -------
        Item
            the random item from the pull
        """
        return random.choices(self.item_list, weights=self.weights, k=1)[0]

    def __str__(self) -> str:
        """Returns a string representation of this Banner object
        self.name
        Price: self.price
        Items:
        .
        .
        .

        Returns
        -------
        str
            String representation of this object
        """
        return (
            self.name
            + "\nPrice: "
            + str(self.price)
            + "\nItems:\n"
            + "\n".join([str(elem) for elem in self.item_list])
        )


class Player:
    """A representation of a player in the gacha game.
    A Player is someone who can use money to purchase pulls from banners,
    receive items from banners, and collect items. Each Player is unique
    and may own different items from each other

    Fields
    ------
    name : str
        the name of the player
    id : str
        the id of the player
        Invariant: must be unique
    items : List[Item]
        the list of items that the player owns
    money : float
        the amount of money that the player owns
    """

    def __init__(self, name: str, id: str, items: List[Item], money: float) -> None:
        """Creates a Player object

        Parameters
        ----------
        name : str
            the name of the player
        id : str
            the id of the player
        items : List[Item]
            the list of items that the player owns
        money : float
            the amount of money that the player owns
        """
        self.name = name
        self.id = id
        self.items = items
        self.money = money

    def add_item(self, item: Item) -> None:
        """Adds an item to the player's inventory

        Parameters
        ----------
        item : Item
            item to add to player inventory

        Returns
        -------
        None
        """
        self.items.append(item)

    def remove_item(self, item: Item) -> bool:
        """Removes an item from the player's inventory

        Parameters
        ----------
        item : Item
            item to add to player inventory

        Returns
        -------
        bool
            True if item is found in inventory, False if not
        """
        try:
            self.items.remove(item)
            return True
        except:
            return False

    def change_money(self, amount: float) -> bool:
        """Adds or removes money from player

        Parameters
        ----------
        amount : float
            the amount to add or remove (positive if add, negative if remove)
            Precondition: if amount removed, leftover amount must be nonnegative

        Returns
        -------
        bool
            True if the amount was able to be added or removed from account
            (does not create a negative money value), False otherwise
        """
        if amount < 0 and self.money + amount < 0:
            return False
        self.money += amount
        return True

    def get_net_worth(self) -> int:
        """Returns the net worth of the player, calculated by the sum of the
        rarities of all of the items they own

        Returns
        -------
        int
            the net worth of the player
        """
        return sum([i.rarity for i in self.items])

    def __str__(self) -> str:
        """Returns a string representation of this Player object
        Money: self.money
        Net worth: self.get_net_worth()
        Top 10 items:
        .
        .
        .
        Top items determined by rarity

        Returns
        -------
        str
            string representation of this object
        """
        return (
            self.name
            + "\n\nMoney: "
            + str(self.money)
            + "\n\nNet worth: "
            + str(self.get_net_worth())
            + "\n\nTop 10 items:\n"
            + "\n".join(
                [
                    str(elem)
                    for elem in sorted(self.items, key=sort_item_key, reverse=True)[:10]
                ]
            )
        )


def sort_item_key(item: Item) -> float:
    """The function used to sort items in a list of items

    Parameters
    ----------
    item : Item
        the item to extract the key from

    Returns
    -------
    float
        the key of the item
    """
    return item.rarity


def get_random_weights(items, key: Callable[[float], float]) -> List[float]:
    """Returns the random weights of the items for the random function

    Parameters
    ----------
    items : List[Item]
        list of items to find weights of
    key : function : float -> float
        function that takes in rarity and returns the drop rate of the item

    Returns
    -------
    List[float]
        the list of weights of the items
    """
    weights = []
    for item in items:
        weights.append(key(item.rarity))
    return weights


"""Collection of loader functions for loading gachapy objects from json files. 
These functions can be used to save and load Controller objects to and from
json files. This should be used to save and load game state of active 
gachapy games.
WARNING: Only use load_controller and save_controller as your top-level loader 
functions. Using the other three helper functions seperately will result in an 
incomplete and uncoupled Controller. Only use if you know what you're doing

These functions operate as the Loader of a gachapy game
"""

from gachapy.objects import *
from gachapy.controller import *
import json


def load_controller(
    items_filename: str, banners_filename: str, players_filename: str
) -> Controller:
    """Creates a Controller object from the specified json files
    TOP LEVEL FUNCTION

    Parameters
    ----------
    items_filename : str
        the path of the items json file
    banners_filename : str
        the path of the banners json file
    players_filename : str
        the path of the players json file

    Returns
    -------
    Controller
        the controller loaded from the specified files
    """
    controller = Controller()
    load_items_from_file(items_filename, controller)
    load_banners_from_file(banners_filename, controller)
    load_players_from_file(players_filename, controller)
    return controller


def save_controller(
    controller: Controller,
    items_filename: str,
    banners_filename: str,
    players_filename: str,
) -> None:
    """Saves the controller in json format into the specified files
    TOP LEVEL FUNCTION

    Parameters
    ----------
    controller : Controller
        the controller to save into files
    items_filename : str
        the path of the items json file
    banners_filename : str
        the path of the banners json file
    players_filename : str
        the path of the players json file

    Returns
    -------
    None
    """
    save_items_to_file(items_filename, controller)
    save_banners_to_file(banners_filename, controller)
    save_players_to_file(players_filename, controller)


def load_items_from_file(filename: str, controller: Controller) -> List[Optional[Item]]:
    """Load items into controller from the specified json file
    DO NOT USE ALONE (unless you know what you're doing)

    Parameters
    ----------
    filename : str
        the path of the json file to load
    controller : Controller
        the controller to load the items into

    Returns
    -------
    List[Optional[Item]]
        the list of items loaded, elements are None if they already exist in controller
    """
    try:
        with open(filename) as f:
            j = json.load(f)
            items_str_list = j["items"]
            return [
                controller.add_new_item(i["name"], i["id"], i["rarity"])
                for i in items_str_list
            ]
    except:
        return []


def load_banners_from_file(
    filename: str, controller: Controller
) -> List[Optional[Banner]]:
    """Load banners into controller from the specified json file
    DO NOT USE ALONE (unless you know what you're doing)

    Parameters
    ----------
    filename : str
        the path of the json file to load
    controller : Controller
        the controller to load the banners into

    Returns
    -------
    List[Optional[Banner]]
        the list of banners loaded, elements are None if they already exist in controller
    """
    try:
        with open(filename) as f:
            j = json.load(f)
            banner_str_list = j["banners"]
            return [
                controller.add_new_banner(
                    i["name"], i["id"], [j["id"] for j in i["items"]], i["price"]
                )
                for i in banner_str_list
            ]
    except:
        return []


def load_players_from_file(
    filename: str, controller: Controller
) -> List[Optional[Player]]:
    """Load players into controller from the specified json file
    DO NOT USE ALONE (unless you know what you're doing)

    Parameters
    ----------
    filename : str
        the path of the json file to load
    controller : Controller
        the controller to load the players into

    Returns
    -------
    List[Optional[Player]]
        the list of players loaded, elements are None if they already exist in controller
    """
    try:
        with open(filename) as f:
            j = json.load(f)
            player_str_list = j["players"]
            return [
                controller.add_new_player(
                    i["name"], i["id"], i["money"], [j["id"] for j in i["items"]]
                )
                for i in player_str_list
            ]
    except:
        return []


def save_items_to_file(filename: str, controller: Controller) -> None:
    """Save items from controller into the specified json file
    DO NOT USE ALONE (unless you know what you're doing)

    Parameters
    ----------
    filename : str
        the path of the json file to save
    controller : Controller
        the controller to load the items from

    Returns
    -------
    None
    """
    items_list = controller.items.values()
    item_dict = {
        "items": [{"name": i.name, "id": i.id, "rarity": i.rarity} for i in items_list]
    }
    with open(filename, "w") as f:
        json.dump(item_dict, f)


def save_banners_to_file(filename: str, controller: Controller) -> None:
    """Save banners from controller into the specified json file
    DO NOT USE ALONE (unless you know what you're doing)

    Parameters
    ----------
    filename : str
        the path of the json file to save
    controller : Controller
        the controller to load the banners from

    Returns
    -------
    None
    """
    banners_list = controller.banners.values()
    banner_dict = {
        "banners": [
            {
                "name": i.name,
                "id": i.id,
                "items": [{"id": j.id} for j in i.item_list],
                "price": i.price,
            }
            for i in banners_list
        ]
    }
    with open(filename, "w") as f:
        json.dump(banner_dict, f)


def save_players_to_file(filename: str, controller: Controller) -> None:
    """Save players from controller into the specified json file
    DO NOT USE ALONE (unless you know what you're doing)

    Parameters
    ----------
    filename : str
        the path of the json file to save
    controller : Controller
        the controller to load the players from

    Returns
    -------
    None
    """
    players_list = controller.players.values()
    player_dict = {
        "players": [
            {
                "name": i.name,
                "id": i.id,
                "money": i.money,
                "items": [{"id": j.id} for j in i.items],
            }
            for i in players_list
        ]
    }
    with open(filename, "w") as f:
        json.dump(player_dict, f)